<el-dialog
  :title=data.{{{widget.widget_id}}}.title
  :size=data.{{{widget.widget_id}}}.size
  :visible.sync="state.{{{widget.widget_id}}}.visible"
>
  <div>
    {{{widget._content}}}
  </div>
</el-dialog>